package org.apollo;

import java.awt.Image;
import java.awt.Toolkit;
import java.awt.event.KeyEvent;
import java.awt.event.KeyListener;
import java.awt.event.WindowEvent;
import java.awt.event.WindowListener;
import java.net.URL;
import java.util.Collection;
import java.util.HashSet;
import java.util.Set;

import javax.swing.SwingUtilities;

import org.apollo.actions.ApolloActions;
import org.apollo.agents.MelodySearch;
import org.apollo.audio.ApolloPlaybackMixer;
import org.apollo.audio.RecordManager;
import org.apollo.audio.SampledAudioManager;
import org.apollo.audio.util.FrameLayoutDaemon;
import org.apollo.audio.util.MultiTrackPlaybackController;
import org.apollo.audio.util.PlaybackClock;
import org.apollo.audio.util.SoundDesk;
import org.apollo.gui.FramePlaybackBarRenderer;
import org.apollo.gui.FrameRenderPasses;
import org.apollo.io.SampledAudioFileImporter;
import org.apollo.util.AudioSystemLog;
import org.expeditee.actions.Actions;
import org.expeditee.gui.Browser;
import org.expeditee.gui.FrameGraphics;
import org.expeditee.importer.FrameDNDTransferHandler;

/**
 * Provides initialization and shutdown services for the Apollo system.
 * 
 * @author Brook Novak
 *
 */
public final class ApolloSystem {
	
	// TODO: Create actual frames
	public static final String SYSTEM_FRAMESET_NAME = "apollosystem";
	public static final String HELP_TOP_FRAMENAME = SYSTEM_FRAMESET_NAME + 2;
	
//	 TODO: How to get good results: collection (moteef) and querry
	// TODO: How to omit indexing on tracks 
	public static final String HELP_MELODYSEARCH_FRAMENAME = SYSTEM_FRAMESET_NAME + 3; 
	
	public static boolean useQualityGraphics = true;
	
	/** Would be nice to seperate outside of this class.. */
	private static boolean isControlDown = false;

	private ApolloSystem() {
	}

	private static boolean hasInitialized = false;
	
	/**
	 * Initializes Apollo mod for expeditee - prepares all subsystems.
	 */
	public static void initialize() {
		
		if (hasInitialized) return;

		AudioSystemLog.println("Initializing...");
		
		SwingUtilities.invokeLater(new Runnable() {
			public void run() {
				
				 try {


					URL url = ClassLoader.getSystemResource("org/apollo/icons/mainicon.png");
					
					if (url != null) {
						 Image img = Toolkit.getDefaultToolkit().getImage(url);
						 Browser._theBrowser.setIconImage(img);
					}
						
	
					 
				 } catch (Exception e) {
					 e.printStackTrace();
				 }
				 
				// Ensure that resources are released before the application is closed.
				Browser._theBrowser.addWindowListener(new WindowListener() {
		
					public void windowActivated(WindowEvent e) {
					}
		
					public void windowClosed(WindowEvent e) {
					}
		
					public void windowClosing(WindowEvent e) {
						ApolloSystem.shutdown();
					}
		
					public void windowDeactivated(WindowEvent e) {
					}
		
					public void windowDeiconified(WindowEvent e) {
					}
		
					public void windowIconified(WindowEvent e) {
					}
		
					public void windowOpened(WindowEvent e) {
					}
				
				});
				
				// Refresh frame for graphical helpers that display / hide when the
				// user presses various keys
				Browser._theBrowser.getContentPane().addKeyListener(new KeyListener() {

					public void keyPressed(KeyEvent e) {
						isControlDown = e.isControlDown();
						if (e.getKeyCode() == KeyEvent.VK_CONTROL) {
							FrameGraphics.refresh(false);
						}
						
					}

					public void keyReleased(KeyEvent e) {
						isControlDown = e.isControlDown();
						if (e.getKeyCode() == KeyEvent.VK_CONTROL) {
							FrameGraphics.refresh(false);
						} else if (e.getKeyCode() == KeyEvent.VK_G && e.isControlDown()) {
							useQualityGraphics = !useQualityGraphics;
						}
					}

					public void keyTyped(KeyEvent e) {
					}
					
				});
			}
		});
		
		// TODO: Have sample rates configurable (remmebr to delete all metadata upen new sample rates).
		// TODO: Fix conversion for me
		 if (SampledAudioManager.PLAYBACK_SAMPLE_RATE != 22050.0f) {
			 AudioSystemLog.print("***BETA-WARNING**: Sample-rate will not give melody searches");
		 }
		 
		// Set title and icon
		 //Browser._theBrowser.etTitle("Apollo");
		 

		 
		System.err.println("WARNING: MISSING MAIN ICON"); // TODO
		
		AudioSystemLog.println("  Preparing sub-systems...");
		
		SampledAudioManager.getInstance();
		
		RecordManager.getInstance();
		
		ApolloPlaybackMixer.getInstance();
		
		FrameLayoutDaemon.getInstance();
		
		FrameRenderPasses.getInstance();
		
		PlaybackClock.getInstance();
		
		FramePlaybackBarRenderer.getInstance();
		
		// Setup for importing audio
		FrameDNDTransferHandler.getInstance().addCustomFileImporter(
				new SampledAudioFileImporter());

		AudioSystemLog.println("  Loading actions and agents...");
		
		// Add apollo actions
		Actions.LoadMethods(ApolloActions.class);
		
		Set<String> agents = new HashSet<String>();
		agents.add(MelodySearch.class.getName());
		
		Collection<String> omitted = Actions.addAgents(agents);
		
		for (String agent : omitted) {
			
			if (agent == null || agent.length() == 0) continue;
			
			String name = agent;
			
			int index = agent.lastIndexOf('.');
			if (index > 0 && agent.length() > (index - 1)) {
				name = agent.substring(index);
			}
			AudioSystemLog.println("  WARNING: Failed to add agent \"" + name + "\"");
			
		}

		AudioSystemLog.println("  Loading banks...");
		SoundDesk.getInstance(); // loads upon creation

		AudioSystemLog.println("Initialized");
		
		hasInitialized = true;
	}
	
	
	

	/**
	 * Releases all resources currently used by the SampledAudioManager.
	 */
	private static void shutdown() {
		
		AudioSystemLog.println("Saving banks...");
		SoundDesk.getInstance().saveMasterMix();
		SoundDesk.getInstance().saveMixes();
		
		AudioSystemLog.println("Releasing resources...");
		
		ApolloPlaybackMixer.getInstance().releaseResources();
		RecordManager.getInstance().releaseResources(); // blocking
		MultiTrackPlaybackController.getInstance().releaseResources();
		
		AudioSystemLog.println("Audio subsystems shutdown");
	}

	/**
	 * @return
	 * 		True if has initialized.
	 */
	public static boolean isInitialized() {
		return hasInitialized;
	}
	
	
	/**
	 * @return True if control is down.
	 */
	public static boolean isControlDown() {
		return isControlDown;
	}




	public static void setControlDown(boolean isControlDown) {
		ApolloSystem.isControlDown = isControlDown;
	}




	/**
	 * The apollo main just ensures that apollo is initialized during startup.
	 * 
	 * This may eventually become redundant once Expeditee implements a plugin system
	 * that allows a plugin / mod to initialize itself at started...
	 * 
	 * @param args
	 */
	public static void main(String[] args) {
		
		// Run expeditee
		Browser.main(args);
		
		// Initialize apollo
		SwingUtilities.invokeLater(new Runnable() {
			public void run() {
				ApolloSystem.initialize();
			}
		});
	}
	
}
